/*
 * $Id: dv1394_audio.cxx,v 1.5 2005/04/03 13:21:11 chutz Exp $
 * vim:sts=2:sw=2:fdm=marker:
 *
 * An attempt to provide a MiniDV driver through the dv1394 kernel module
 *
 * Requirements:
 * - dv1394 kernel module loaded and running
 * - libdv (tested with 0.102)
 *
 * Author: Georgi Georgiev <chutz@gg3.net>
 *
 */

#error This thing does not work at all
This_thing_does_not_work_at_all();

#pragma implementation "dv1394_audio.h"

#include "dv1394_audio.h"

PCREATE_SOUND_PLUGIN(DV, PSoundChannelDV1394)

#define BUF_FRAMES 1
#define PAL_FRAME_SIZE 150000
#define NTSC_FRAME_SIZE 120000
#define MAX_FRAME_SIZE PAL_FRAME_SIZE

static PMutex mutex;
static PDictionary<PString, PString> *dico;
static BYTE rawFrameBuffer[BUF_FRAMES][MAX_FRAME_SIZE];
static int frameToWrite;
static int frameToRead;
#if DEBUG >= 2
static int frameDiff;
static int frameTotal;
#endif

// InputDVGrabber /* {{{ */
//

InputDVGrabber::InputDVGrabber() : PThread(1000, NoAutoDeleteThread) {
  grabberExitFlag = FALSE;
  fd = -1;
}

void InputDVGrabber::Main() {
  while (1) {
    grabberMutex.Wait();
    if (grabberExitFlag == TRUE) {
      grabberMutex.Signal();
      break;
    }
    grabberMutex.Signal();
    ReadAnotherFrame();
  }
}

void InputDVGrabber::ReadAnotherFrame() {
  if (fd < 0) {
    Sleep(10);
    return;
  }

  BYTE buffer[MAX_FRAME_SIZE];
  BYTE * buffer_end = buffer;
  // PAL or NTSC?
  buffer_end += ::read(fd, buffer, 4);
  int thisFrameSize;
  if (buffer[2] & 0x80) {
    thisFrameSize = PAL_FRAME_SIZE;
#if DEBUG >= 2
    ::cerr << "Detected PAL format" << ::endl;
#endif
  } else {
    thisFrameSize = NTSC_FRAME_SIZE;
#if DEBUG >= 2
    ::cerr << "Detected NTSC format" << ::endl;
#endif
  }
  while (buffer_end - buffer < thisFrameSize) {
    buffer_end += ::read(fd, buffer_end, thisFrameSize - (buffer_end - buffer));
  }
  mutex.Wait();
  memcpy(rawFrameBuffer[frameToWrite], buffer, thisFrameSize);
#if ! BUF_FRAMES > 1
  frameToRead = 0;
#else
  if (frameToRead < 0)             { frameToRead = 0; }
  frameToWrite++;
#if DEBUG >= 2
  frameDiff++;
  frameTotal++;
  PTime t;
  ::cerr << "WRITING    "
    << t.GetHour() << ":" << t.GetMinute() << ":" << t.GetSecond() << "." << t.GetMicrosecond() << "\t"
    << "DIFF " << frameDiff << "/" << frameTotal << "\t" << (double)frameDiff/(double)frameTotal*100.0 << "%" << endl;
#endif
  if (frameToWrite >= BUF_FRAMES)  { frameToWrite = 0; }
  if (frameToRead == frameToWrite) { frameToRead++; }
  if (frameToRead >= BUF_FRAMES)   { frameToRead = 0; }
#endif
  mutex.Signal();
};

void InputDVGrabber::Stop() {
  grabberMutex.Wait();
  grabberExitFlag = TRUE;
  grabberMutex.Signal();
}
/* }}} */
// PSoundChannelDV1394  /* {{{ */
///////////////////////////////////////////////////////////////////////////////

PSoundChannelDV1394::PSoundChannelDV1394() // {{{
{
  fd = -1;
  dv_decoder = NULL;
  grabber    = NULL;
  myNumChannels	= 2;
  mySampleRate	= 32000;
  myBitsPerSample = 16;
  Initialize();
}
// }}}
PSoundChannelDV1394::~PSoundChannelDV1394()/*{{{*/
{
  Close();
}
/*}}}*/
PStringArray PSoundChannelDV1394::GetDeviceNames ( Directions dir) /*{{{*/
{
  PStringArray Result;
  
  /*
   * Looking for devices created by the dv1394 kernel module:
   * - devfs used to create these in /dev/..../NTSC/{in,out} or something
   * - udev-056 and linux-kernel 2.6.12-rc1 created a node at /dev/dv1394-0 a a
   * Make this thing work with my setup, apologize and support the rest later
   *
   * TODO: Borrow the ReadDirectory code from v4l2 maybe?
   */

  for (int i=0; ; i++) {
    PString *devname = new PString("/dev/dv1394/" + (PString) i);
    // PString ufname = "Camera "+ (PString) i + " (dv1394/" + (PString) i + ")";
    PString ufname = *devname;
    struct stat dummy;
    if (::stat(*devname, &dummy) != 0) {
      break;
    }
    mutex.Wait();
    if(dico == NULL)
    {
      dico = new PDictionary<PString, PString>;
    }
    dico->SetAt(ufname, devname);
    mutex.Signal();
    Result.AppendString(ufname);
  }
  return Result;
}
/*}}}*/
PString PSoundChannelDV1394::GetDefaultDevice ( Directions dir) /*{{{*/
{
  PStringArray devices;
  devices = GetDeviceNames(dir);
  return devices[0];
};
  
/*}}}*/
BOOL PSoundChannelDV1394::Open( /*{{{*/
    const PString & _device,
    Directions _dir,
    unsigned _numChannels,
    unsigned _sampleRate,
    unsigned _bitsPerSample)
{
  PString * devName;

  if (IsOpen()) Close();

  mutex.Wait();
  if(dico != NULL)
    devName = dico->GetAt(_device);
  else
    return FALSE;
  mutex.Signal();

  fd = ::open(*devName, O_RDONLY);
  if (fd == -1) {
    PTRACE(3, "Open() failed");
    return FALSE;
  }
  if (!Start()) {
    Stop();
  }
  return TRUE;
}
/*}}}*/
BOOL PSoundChannelDV1394::IsOpen()/*{{{*/
{
  return fd != -1;
}
/*}}}*/
BOOL PSoundChannelDV1394::Read (void * buf, PINDEX len) {/*{{{*/
  memset ((char *) buf, 0, len);

  PINDEX copied = 0;

  while (len < copied) {
    if ( (len - copied) > (decBufEnd - decBufPos)) {
      // What we need is not in the buffer yet
      memcpy((BYTE *)buf + copied, decodedBuffer + decBufPos, decBufEnd - decBufPos);
      copied += decBufEnd - decBufPos;
      decBufPos = decBufEnd;
      // Refill the buffer;
      FillDecodedBuffer();
    } else {
      // The buffer has all the required data
      memcpy((BYTE *)buf + copied, decodedBuffer + decBufPos, len - copied);
      decBufPos += len - copied;
      copied = len; // I know, just for clarity
      break;
    }
  }

  return TRUE;
}
/*}}}*/
BOOL PSoundChannelDV1394::SetFormat(/*{{{*/
    unsigned numChannels,
    unsigned sampleRate,
    unsigned bitsPerSample
    )
{
  myNumChannels   = numChannels;
  mySampleRate    = sampleRate;
  myBitsPerSample = bitsPerSample;
  Initialize();
  return TRUE;
};
/*}}}*/
BOOL PSoundChannelDV1394::SetBuffers(PINDEX size, PINDEX count) {/*{{{*/
  bufSize = size;
  bufCount = count;
  return TRUE;
}
/*}}}*/
BOOL PSoundChannelDV1394::GetBuffers(PINDEX & size, PINDEX & count) {/*{{{*/
  size = bufSize;
  count = bufCount;
  return TRUE;
}
/*}}}*/
BOOL PSoundChannelDV1394::IsStarted() /*{{{*/
{
  return grabber != NULL;
}
/*}}}*/
BOOL PSoundChannelDV1394::Initialize() {/*{{{*/
  decBufEnd	= 0;
  decBufPos	= 0;
  return TRUE;
}
/*}}}*/
unsigned PSoundChannelDV1394::GetChannels() const /*{{{*/
{
  return myNumChannels;
}
/*}}}*/
unsigned PSoundChannelDV1394::GetSampleRate() const /*{{{*/
{
  return mySampleRate;
}
/*}}}*/
unsigned PSoundChannelDV1394::GetSampleSize() const /*{{{*/
{
  return myBitsPerSample;
}
/*}}}*/

BOOL PSoundChannelDV1394::FillDecodedBuffer() {/*{{{*/
  if (!IsStarted()) return FALSE;

  /*
  int read_frames = 0;
  BYTE * capture_buffer;
  BYTE * capture_buffer_end = capture_buffer;
  */
  mutex.Wait();
  if (frameToRead < 0) {
    mutex.Signal();
    return FALSE;
  }
  BYTE rawFrame[MAX_FRAME_SIZE];
  memcpy(rawFrame, rawFrameBuffer[frameToRead], MAX_FRAME_SIZE);
#if BUF_FRAMES > 1
  frameToRead++;
  if (frameToRead >= BUF_FRAMES) { frameToRead = 0; }
#ifdef DEBUG
  PTime t;
  frameDiff--;
  cout << "READING    "
    << t.GetHour() << ":" << t.GetMinute() << ":" << t.GetSecond() << "." << t.GetMicrosecond() << "\t" <<
    "DIFF " << frameDiff << "/" << frameTotal << "\t" << (double)frameDiff/(double)frameTotal*100.0 << "%" << endl;
#endif
#endif
  mutex.Signal();
  
  dv_decoder_t *dv;
  dv = dv_decoder_new(TRUE, FALSE, FALSE);
  /* possible quality values from dv_types.h:
   * DV_QUALITY_
   *            FASTEST : grayscale, thumbnail-size image
   *            AC_1
   *            AC_2
   *            DC
   *            COLOR
   *            AC_MASK
   *            BEST    : COLOR | AC_2
   */
  dv->quality = DV_QUALITY_BEST;
  if(dv_parse_header(dv, rawFrame) < 0) {
    PTRACE(3, "cannot parse dv frame header");
    return FALSE;
  }
  int16_t *outbuf[4];
  
  dv_decode_full_audio(dv, rawFrame, outbuf);

  // Rescaling... resampling I mean. *MANUALLY* (I wish I knew how to do this
  // properly

  int dvSamples       = dv_get_num_samples(dv);
  int dvChannels      = dv_get_num_channels(dv);
  /*
  int dvSampleRate    = dv_get_frequency(dv);
  */

  for (int sample=0; sample < dvSamples; sample++) {
    for (int chan=0; chan < myNumChannels; chan++) {
      int16_t val;
      if (myNumChannels <= dvChannels) {
	val = outbuf[chan][sample];
      } else {
	val = outbuf[chan * myNumChannels/dvChannels][sample];
      }
      switch (myBitsPerSample) {
	case 8:
	  decodedBuffer[((sample * myNumChannels) + chan)] = val >> 8;
	  break;
	case 16:
	default:
	  decodedBuffer[(sample * myNumChannels + chan) + 0] = val >> 8;
	  decodedBuffer[(sample * myNumChannels + chan) + 1] = val & 0xFF;
      };
    };
  };

  decBufPos = 0;
  decBufEnd = myNumChannels * dvSamples * myBitsPerSample;

  return TRUE;
}
/*}}}*/

BOOL PSoundChannelDV1394::Start() /*{{{*/
{
  if (!IsOpen()) return FALSE;
  if (!IsStarted()) {
    mutex.Wait();
    frameToWrite = 0;
#if BUF_FRAMES > 1
    frameToRead = -1;
    frameDiff   = 0;
    frameTotal  = 0;
#else
    frameToRead = 0;
#endif
    grabber = new InputDVGrabber();
    grabber->fd = fd;
    grabber->Resume();
    mutex.Signal();
  }
  return TRUE;
}
/*}}}*/
BOOL PSoundChannelDV1394::Stop()/*{{{*/
{
  if (IsStarted()) {
    grabber->Stop();
    grabber->WaitForTermination();
    delete grabber;
    grabber = NULL;
  }
  return TRUE;
}
/*}}}*/
BOOL PSoundChannelDV1394::Close()/*{{{*/
{
  PTRACE(3, "Close()");
  if (IsStarted()) {
    Stop();
    if(::close(fd) == 0) {
      fd = -1;
      return TRUE;
    } else {
      return FALSE;
    }
  }
  return FALSE;
}
/*}}}*/
// /* }}} */
// End Of File ///////////////////////////////////////////////////////////////
