/*
 * $Id: dv1394_audio.h,v 1.3 2005/04/03 11:01:19 chutz Exp $
 * vim:sts=2:sw=2:fdm=marker:
 *
 * audioinput_dv.h
 * Provide audio input for MiniDV cameras. Requires the dv1394 kernel module.
 *
 * Author: Georgi Georgiev <chutz@gg3.net>
 *
 */


#ifdef __GNUC__
#pragma interface
#endif

#include <libdv/dv.h>
#include <sys/stat.h>

#include <ptlib.h>
#include <ptlib/sound.h>
#include <ptlib/file.h>
#if !P_USE_INLINES
#include <ptlib/contain.inl>
#endif

#define DV_AUDIO_MAX_CHANNELS 4

/*
 * The dv1394 module creates a device that provides DV data. However, reading
 * *on-request* is no good, since frames tend to get lost because it is like this:
 *   - the reader reads a frame (frame N)
 *   - the reader returns that frame, meanwhile frame N+1 is already started on
 *   the device, but nobody reads it.
 *   - next time the reader tries to read a frame, N+1 is missed, so it waits
 *   for frame N+2
 *
 *   At least this is what the problem was with the old 1394AVC implementation
 *
 *   Trying to avoid this problem by introducing a thread that reads continuously.
 */

class InputDVGrabber : public PThread/*{{{*/
{
  PCLASSINFO(InputDVGrabber, PThread);
  public:
    InputDVGrabber();
    void	Main();
    void	ReadAnotherFrame();
    void	Stop();

    int		fd;
  protected:
    PMutex	grabberMutex;
    BOOL	grabberExitFlag;
};
/*}}}*/
class PSoundChannelDV1394 : public PSoundChannel/*{{{*/
{
  PCLASSINFO(PSoundChannelDV1394, PSoundChannel)

  public:
    PSoundChannelDV1394();
    ~PSoundChannelDV1394();
    static PStringArray GetDeviceNames(PSoundChannel::Directions);
    static PString GetDefaultDevice(PSoundChannel::Directions);
    BOOL Open(
	const PString & deviceName,   /// Device name to open
	Directions dir,
	unsigned _numChannels,
	unsigned _sampleRate,
	unsigned _bitsPerSample
	);

    BOOL Close();
    BOOL Read(void * buf, PINDEX len);
    BOOL SetFormat(
	unsigned numChannels,
	unsigned sampleRate,
	unsigned bitsPerSample
	);
    unsigned GetChannels() const;
    unsigned GetSampleRate() const;
    unsigned GetSampleSize() const;
    BOOL SetBuffers(PINDEX size, PINDEX count = 2);
    BOOL GetBuffers(PINDEX & size, PINDEX & count);
    BOOL RecordSound(PSound & sound);
    BOOL RecordFile(const PFilePath & filename);
    BOOL StartRecording();
    BOOL IsRecordBufferFull();
    BOOL AreAllRecordBuffersFull();
    BOOL WaitForRecordBufferFull();
    BOOL WaitForAllRecordBuffersFull();
    BOOL Abort();
    BOOL SetVolume (unsigned);
    BOOL GetVolume (unsigned &);
    BOOL IsOpen();

  protected:

    int frameBytes; 
    dv_decoder_t *	dv_decoder;
    int		fd;
    InputDVGrabber	* grabber;

  private:
    int mySampleRate;
    int myBitsPerSample;
    int myNumChannels;
    BYTE decodedBuffer[DV_AUDIO_MAX_CHANNELS * DV_AUDIO_MAX_SAMPLES * 2];
    int decBufPos;
    int decBufEnd;
    int bufSize;
    int bufCount;

    BOOL Initialize();
    BOOL Start();
    BOOL Stop();
    BOOL FillDecodedBuffer();
    BOOL IsStarted();
};
/*}}}*/
